@@ -27,10 +27,9 @@ from kornia.filters import filter2D
 
 import torchvision
 from torchvision import transforms
+from stylegan2_pytorch.version import __version__
 from stylegan2_pytorch.diff_augment import DiffAugment
 
-from pytorch_fid import fid_score
-
 from vector_quantize_pytorch import VectorQuantize
 from linear_attention_transformer import ImageLinearAttention
 
@@ -43,6 +42,8 @@ try:
 except:
     APEX_AVAILABLE = False
 
+import aim
+
 assert torch.cuda.is_available(), 'You need to have an Nvidia GPU with CUDA installed.'
 
 num_cores = multiprocessing.cpu_count()
@@ -51,7 +52,6 @@ num_cores = multiprocessing.cpu_count()
 
 EXTS = ['jpg', 'jpeg', 'png']
 EPS = 1e-8
-EVALUATE_EVERY = 1000
 CALC_FID_NUM_IMAGES = 12800
 
 # helper classes
@@ -330,7 +330,7 @@ class AugWrapper(nn.Module):
             images = DiffAugment(images, types=types)
 
         if detach:
-            images.detach_()
+            images = images.detach()
 
         return self.D(images)
 
@@ -558,7 +558,7 @@ class Discriminator(nn.Module):
         num_init_filters = 3 if not transparent else 4
 
         blocks = []
-        filters = [num_init_filters] + [(64) * (2 ** i) for i in range(num_layers + 1)]
+        filters = [num_init_filters] + [(network_capacity * 4) * (2 ** i) for i in range(num_layers + 1)]
 
         set_fmap_max = partial(min, fmap_max)
         filters = list(map(set_fmap_max, filters))
@@ -686,18 +686,63 @@ class StyleGAN2(nn.Module):
         return x
 
 class Trainer():
-    def __init__(self, name, results_dir, models_dir, image_size, network_capacity, transparent = False, batch_size = 4, mixed_prob = 0.9, gradient_accumulate_every=1, lr = 2e-4, lr_mlp = 1., ttur_mult = 2, rel_disc_loss = False, num_workers = None, save_every = 1000, trunc_psi = 0.6, fp16 = False, cl_reg = False, fq_layers = [], fq_dict_size = 256, attn_layers = [], no_const = False, aug_prob = 0., aug_types = ['translation', 'cutout'], top_k_training = False, generator_top_k_gamma = 0.99, generator_top_k_frac = 0.5, dataset_aug_prob = 0., calculate_fid_every = None, is_ddp = False, rank = 0, world_size = 1, *args, **kwargs):
+    def __init__(
+        self,
+        name = 'default',
+        results_dir = 'results',
+        models_dir = 'models',
+        base_dir = './',
+        image_size = 128,
+        network_capacity = 16,
+        fmap_max = 512,
+        transparent = False,
+        batch_size = 4,
+        mixed_prob = 0.9,
+        gradient_accumulate_every=1,
+        lr = 2e-4,
+        lr_mlp = 1.,
+        ttur_mult = 2,
+        rel_disc_loss = False,
+        num_workers = None,
+        save_every = 1000,
+        evaluate_every = 1000,
+        num_image_tiles = 8,
+        trunc_psi = 0.6,
+        fp16 = False,
+        cl_reg = False,
+        fq_layers = [],
+        fq_dict_size = 256,
+        attn_layers = [],
+        no_const = False,
+        aug_prob = 0.,
+        aug_types = ['translation', 'cutout'],
+        top_k_training = False,
+        generator_top_k_gamma = 0.99,
+        generator_top_k_frac = 0.5,
+        dataset_aug_prob = 0.,
+        calculate_fid_every = None,
+        is_ddp = False,
+        rank = 0,
+        world_size = 1,
+        log = False,
+        *args,
+        **kwargs
+    ):
         self.GAN_params = [args, kwargs]
         self.GAN = None
 
         self.name = name
-        self.results_dir = Path(results_dir)
-        self.models_dir = Path(models_dir)
+
+        base_dir = Path(base_dir)
+        self.base_dir = base_dir
+        self.results_dir = base_dir / results_dir
+        self.models_dir = base_dir / models_dir
         self.config_path = self.models_dir / name / '.config.json'
 
         assert log2(image_size).is_integer(), 'image size must be a power of 2 (64, 128, 256, 512, 1024)'
         self.image_size = image_size
         self.network_capacity = network_capacity
+        self.fmap_max = fmap_max
         self.transparent = transparent
 
         self.fq_layers = cast_list(fq_layers)
@@ -718,6 +763,8 @@ class Trainer():
         self.num_workers = num_workers
         self.mixed_prob = mixed_prob
 
+        self.num_image_tiles = num_image_tiles
+        self.evaluate_every = evaluate_every
         self.save_every = save_every
         self.steps = 0
 
@@ -758,6 +805,8 @@ class Trainer():
         self.rank = rank
         self.world_size = world_size
 
+        self.logger = aim.Session(experiment=name) if log else None
+
     @property
     def image_extension(self):
         return 'jpg' if not self.transparent else 'png'
@@ -765,10 +814,14 @@ class Trainer():
     @property
     def checkpoint_num(self):
         return floor(self.steps // self.save_every)
-    
+
+    @property
+    def hparams(self):
+        return {'image_size': self.image_size, 'network_capacity': self.network_capacity}
+        
     def init_GAN(self):
         args, kwargs = self.GAN_params
-        self.GAN = StyleGAN2(lr = self.lr, lr_mlp = self.lr_mlp, ttur_mult = self.ttur_mult, image_size = self.image_size, network_capacity = self.network_capacity, transparent = self.transparent, fq_layers = self.fq_layers, fq_dict_size = self.fq_dict_size, attn_layers = self.attn_layers, fp16 = self.fp16, cl_reg = self.cl_reg, no_const = self.no_const, rank = self.rank, *args, **kwargs)
+        self.GAN = StyleGAN2(lr = self.lr, lr_mlp = self.lr_mlp, ttur_mult = self.ttur_mult, image_size = self.image_size, network_capacity = self.network_capacity, fmap_max = self.fmap_max, transparent = self.transparent, fq_layers = self.fq_layers, fq_dict_size = self.fq_dict_size, attn_layers = self.attn_layers, fp16 = self.fp16, cl_reg = self.cl_reg, no_const = self.no_const, rank = self.rank, *args, **kwargs)
 
         if self.is_ddp:
             ddp_kwargs = {'device_ids': [self.rank]}
@@ -777,6 +830,9 @@ class Trainer():
             self.D_ddp = DDP(self.GAN.D, **ddp_kwargs)
             self.D_aug_ddp = DDP(self.GAN.D_aug, **ddp_kwargs)
 
+        if exists(self.logger):
+            self.logger.set_params(self.hparams)
+
     def write_config(self):
         self.config_path.write_text(json.dumps(self.config()))
 
@@ -787,6 +843,7 @@ class Trainer():
         self.transparent = config['transparent']
         self.fq_layers = config['fq_layers']
         self.fq_dict_size = config['fq_dict_size']
+        self.fmap_max = config.pop('fmap_max', 512)
         self.attn_layers = config.pop('attn_layers', [])
         self.no_const = config.pop('no_const', False)
         del self.GAN
@@ -897,6 +954,7 @@ class Trainer():
             if apply_gradient_penalty:
                 gp = gradient_penalty(image_batch, real_output)
                 self.last_gp_loss = gp.clone().detach().item()
+                self.track(self.last_gp_loss, 'GP')
                 disc_loss = disc_loss + gp
 
             disc_loss = disc_loss / self.gradient_accumulate_every
@@ -906,6 +964,8 @@ class Trainer():
             total_disc_loss += divergence.detach().item() / self.gradient_accumulate_every
 
         self.d_loss = float(total_disc_loss)
+        self.track(self.d_loss, 'D')
+
         self.GAN.D_opt.step()
 
         # train generator
@@ -950,12 +1010,15 @@ class Trainer():
             total_gen_loss += loss.detach().item() / self.gradient_accumulate_every
 
         self.g_loss = float(total_gen_loss)
+        self.track(self.g_loss, 'G')
+
         self.GAN.G_opt.step()
 
         # calculate moving averages
 
         if apply_path_penalty and not np.isnan(avg_pl_length):
             self.pl_mean = self.pl_length_ma.update_average(self.pl_mean, avg_pl_length)
+            self.track(self.pl_mean, 'PL')
 
         if self.is_main and self.steps % 10 == 0 and self.steps > 20000:
             self.GAN.EMA()
@@ -976,8 +1039,8 @@ class Trainer():
             if self.steps % self.save_every == 0:
                 self.save(self.checkpoint_num)
 
-            if self.steps % EVALUATE_EVERY == 0 or (self.steps % 100 == 0 and self.steps < 2500):
-                self.evaluate(floor(self.steps / EVALUATE_EVERY))
+            if self.steps % self.evaluate_every == 0 or (self.steps % 100 == 0 and self.steps < 2500):
+                self.evaluate(floor(self.steps / self.evaluate_every))
 
             if exists(self.calculate_fid_every) and self.steps % self.calculate_fid_every == 0 and self.steps != 0:
                 num_batches = math.ceil(CALC_FID_NUM_IMAGES / self.batch_size)
@@ -991,10 +1054,10 @@ class Trainer():
         self.av = None
 
     @torch.no_grad()
-    def evaluate(self, num = 0, num_image_tiles = 8, trunc = 1.0):
+    def evaluate(self, num = 0, trunc = 1.0):
         self.GAN.eval()
         ext = self.image_extension
-        num_rows = num_image_tiles
+        num_rows = self.num_image_tiles
     
         latent_dim = self.GAN.G.latent_dim
         image_size = self.GAN.G.image_size
@@ -1037,6 +1100,7 @@ class Trainer():
 
     @torch.no_grad()
     def calculate_fid(self, num_batches):
+        from pytorch_fid import fid_score
         torch.cuda.empty_cache()
 
         real_path = str(self.results_dir / self.name / 'fid_real') + '/'
@@ -1075,23 +1139,34 @@ class Trainer():
         return fid_score.calculate_fid_given_paths([real_path, fake_path], 256, True, 2048)
 
     @torch.no_grad()
-    def generate_truncated(self, S, G, style, noi, trunc_psi = 0.75, num_image_tiles = 8):
-        latent_dim = G.latent_dim
+    def truncate_style(self, tensor, trunc_psi = 0.75):
+        S = self.GAN.S
+        batch_size = self.batch_size
+        latent_dim = self.GAN.G.latent_dim
 
         if not exists(self.av):
             z = noise(2000, latent_dim, device=self.rank)
-            samples = evaluate_in_chunks(self.batch_size, S, z).cpu().numpy()
+            samples = evaluate_in_chunks(batch_size, S, z).cpu().numpy()
             self.av = np.mean(samples, axis = 0)
             self.av = np.expand_dims(self.av, axis = 0)
-            
-        w_space = []
-        for tensor, num_layers in style:
-            tmp = S(tensor)
-            av_torch = torch.from_numpy(self.av).cuda(self.rank)
-            tmp = trunc_psi * (tmp - av_torch) + av_torch
-            w_space.append((tmp, num_layers))
 
-        w_styles = styles_def_to_tensor(w_space)
+        av_torch = torch.from_numpy(self.av).cuda(self.rank)
+        tensor = trunc_psi * (tensor - av_torch) + av_torch
+        return tensor
+
+    @torch.no_grad()
+    def truncate_style_defs(self, w, trunc_psi = 0.75):
+        w_space = []
+        for tensor, num_layers in w:
+            tensor = self.truncate_style(tensor, trunc_psi = trunc_psi)            
+            w_space.append((tensor, num_layers))
+        return w_space
+
+    @torch.no_grad()
+    def generate_truncated(self, S, G, style, noi, trunc_psi = 0.75, num_image_tiles = 8):
+        w = map(lambda t: (S(t[0]), t[1]), style)
+        w_truncated = self.truncate_style_defs(w, trunc_psi = trunc_psi)
+        w_styles = styles_def_to_tensor(w_truncated)
         generated_images = evaluate_in_chunks(self.batch_size, G, w_styles, noi)
         return generated_images.clamp_(0., 1.)
 
@@ -1150,6 +1225,11 @@ class Trainer():
         log = ' | '.join(map(lambda n: f'{n[0]}: {n[1]:.2f}', data))
         print(log)
 
+    def track(self, value, name):
+        if not exists(self.logger):
+            return
+        self.logger.track(value, name = name)
+
     def model_name(self, num):
         return str(self.models_dir / self.name / f'model_{num}.pt')
 
@@ -1158,13 +1238,16 @@ class Trainer():
         (self.models_dir / self.name).mkdir(parents=True, exist_ok=True)
 
     def clear(self):
-        rmtree(f'./models/{self.name}', True)
-        rmtree(f'./results/{self.name}', True)
+        rmtree(str(self.models_dir / self.name), True)
+        rmtree(str(self.results_dir / self.name), True)
         rmtree(str(self.config_path), True)
         self.init_folders()
 
     def save(self, num):
-        save_data = {'GAN': self.GAN.state_dict()}
+        save_data = {
+            'GAN': self.GAN.state_dict(),
+            'version': __version__
+        }
 
         if self.GAN.fp16:
             save_data['amp'] = amp.state_dict()
@@ -1188,11 +1271,38 @@ class Trainer():
 
         load_data = torch.load(self.model_name(name))
 
-        # make backwards compatible
-        if 'GAN' not in load_data:
-            load_data = {'GAN': load_data}
-
-        self.GAN.load_state_dict(load_data['GAN'])
+        if 'version' in load_data:
+            print(f"loading from version {load_data['version']}")
 
+        try:
+            self.GAN.load_state_dict(load_data['GAN'])
+        except Exception as e:
+            print('unable to load save model. please try downgrading the package to the version specified by the saved model')
+            raise e
         if self.GAN.fp16 and 'amp' in load_data:
             amp.load_state_dict(load_data['amp'])
+
+class ModelLoader:
+    def __init__(self, *, base_dir, name = 'default', load_from = -1):
+        self.model = Trainer(name = name, base_dir = base_dir)
+        self.model.load(load_from)
+
+    def noise_to_styles(self, noise, trunc_psi = None):
+        noise = noise.cuda()
+        w = self.model.GAN.S(noise)
+        if exists(trunc_psi):
+            w = self.model.truncate_style(w)
+        return w
+
+    def styles_to_images(self, w):
+        batch_size, *_ = w.shape
+        num_layers = self.model.GAN.G.num_layers
+        image_size = self.model.image_size
+        w_def = [(w, num_layers)]
+
+        w_tensors = styles_def_to_tensor(w_def)
+        noise = image_noise(batch_size, image_size, device = 0)
+
+        images = self.model.GAN.G(w_tensors, noise)
+        images.clamp_(0., 1.)
+        return images
